#include <PSX/GPU/gpu.hpp>
#include <PSX/GPU/make_gpu_primitives.hpp>
#include <FontWriter/font_writer.hpp>
#include <stdio.h>

namespace JabyEngine {
    static const char* simple_itoa(char (&buffer)[32], int value, int base) {
        const bool is_neg = value < 0;
        int        i      = 32 - 2;

        if(is_neg) {
            value *= -1;
        }
    
        for(; value && i; i--, value /= base) {
            char chr = '0' + value%base;
            if(chr > '9') {
                chr += ('A' - '9');
            }
            buffer[i] = chr;
        }

        buffer[32 - 1] = '\0';
        if(is_neg) {
            buffer[i] = '-';
            return &buffer[i];
        }
        return &buffer[i+1];
    }

    void FontWriter :: setup(const FontBufferInfo& buffer_info, const SimpleTIM& vram_dst, const GPU::SizeI16& font_size) {
        this->prim_buffer   = buffer_info;
        this->tex_page      = Make::TexPage(vram_dst.get_texture_position(), GPU::TexturePageColor::$4bit).linked();
        this->cur_primitive = buffer_info.double_buffer[GPU::Display::current_id];
        
        for(size_t buffer_id = 0; buffer_id < 2; buffer_id++) {
            for(size_t buffer_element_id = 0; buffer_element_id < buffer_info.single_buffer_length; buffer_element_id++) {
                this->prim_buffer.double_buffer[buffer_id][buffer_element_id] = Make::SPRT(
                    Make::AreaI16(Make::PositionI16(0, 0), font_size),
                    Make::OffsetPageWithClut(Make::PageOffset(0, 0), Make::PageClut(vram_dst.get_clut_position()))
                ).linked();
            }
        }
    }

    void FontWriter :: write(State& state, const char* str, GPU::Color24 color, Wiggle* wiggle, va_list b) {
        const auto* primitive_end = &this->prim_buffer.double_buffer[GPU::Display::current_id][this->prim_buffer.single_buffer_length];
        const auto  font_size     = this->prim_buffer.double_buffer[0][0]->size;
        const auto  row_count     = 256/font_size.width;
        const auto  push_char     = [&](State& state, FontPrimitive& primitive, char cur_char) -> bool {
            if(&primitive >= primitive_end) {
                return false;
            }

            primitive->position = state.pos;
            if(wiggle) {
                const auto [off_x, off_y] =  (*wiggle)[(state.wiggle_count++)&0x7];
                primitive->position.move(off_x, off_y);
            }
            primitive->tex_offset = GPU::PageOffset::from_tile_id(cur_char - '!', row_count, font_size);
            primitive->color      = color;
            primitive.concat(*this->cur_primitive);
            state.pos.move(primitive->size.width, 0);
            return true;
        };
        const auto  old_x    = state.pos.x;
        const char* prev_str = nullptr;
        char        buffer[32]; 

        while(true) {
            auto cur_char = *str++;
            switch(cur_char) {
                case '\0':
                    if(prev_str) {
                        str = prev_str;
                        prev_str = nullptr;
                        continue;
                    }
                    return;

                case '\n':
                    state.pos.x  = old_x;
                    state.pos.y += font_size.height;
                    continue;

                case ' ':
                    state.pos.x += font_size.width;
                    continue;

                case '%':
                    switch(*str) {
                        case 'i':
                            prev_str = str + 1;
                            str = simple_itoa(buffer, -505, 10);
                            continue;

                        case '%':
                            str++;
                            break;
                    }

                default:
                    auto& primitive = *this->cur_primitive++;
                    if(!push_char(state, primitive, cur_char)) {
                        return;
                    }
            }   
        }
    }

    void FontWriter :: render() {
        const auto update_id = GPU::Display::current_id;
        const auto render_id = update_id ^ 1;

        this->cur_primitive -= 1;
        if(this->cur_primitive >= this->prim_buffer.double_buffer[render_id]) {
            this->cur_primitive->terminate();
            this->tex_page.concat(*this->prim_buffer.double_buffer[render_id]);

            GPU::render(this->tex_page);
        }

        this->cur_primitive = this->prim_buffer.double_buffer[update_id];
    }
}